/**
 * b64_decode - Base64 decode
 * @src: Data to be decoded
 * @len: Length of the data to be decoded
 * @out_len: Pointer to output length variable, or %NULL if not used
 * Returns: Allocated buffer of out_len bytes of decoded data,
 * or %NULL on failure
 *
 * Caller is responsible for freeing the returned buffer.
 */
unsigned char* b64_decode(const unsigned char *src, size_t len, size_t *out_len)
{
    unsigned char dtable[256], *out, *pos, block[4], tmp;
    size_t i, count, olen;
    int pad = 0;

    memset(dtable, 0x80, 256);
    for (i = 0; i < sizeof(base64_table) - 1; i++)
        dtable[base64_table[i]] = (unsigned char) i;
    dtable['='] = 0;

    count = 0;
    for (i = 0; i < len; i++) 
        {
        if (dtable[src[i]] != 0x80)
            count++;
	}

    if (count == 0 || count % 4)
        return NULL;

    olen = count / 4 * 3;
    pos = out = malloc(olen);
    if (out == NULL)
        return NULL;

    count = 0;
    for (i = 0; i < len; i++) 
        {
        tmp = dtable[src[i]];
        if (tmp == 0x80)
            continue;

        if (src[i] == '=')
            pad++;
        block[count] = tmp;
        count++;
        if (count == 4) 
            {
            *pos++ = (block[0] << 2) | (block[1] >> 4);
            *pos++ = (block[1] << 4) | (block[2] >> 2);
            *pos++ = (block[2] << 6) | block[3];
            count = 0;
            if (pad) 
                {
                if (pad == 1)
                    pos--;
                else if (pad == 2)
                    pos -= 2;
                else {
                    /* Invalid padding */
                    free(out);
                    return NULL;
                    }
                break;
                }
            }
	}
	
    size_t _out_len = pos - out;

    if(out_len)
        *out_len = _out_len;
	
    // Sometimes _out_len is different than strlen(out)
    // Because _out_len holds correct value
    // let's apply this dirty hack to fix output length

    if(_out_len != strlen((const char*) out)) 
        {
        unsigned char* newout = realloc(out, _out_len + 1);
        if(newout) 
            {
            newout[_out_len] = '\0';
            out = newout;
            } 
        else 
            {
            free(out);
            out = NULL;
            }
	}

    return out;
}

/**
 * b64_encode - Base64 encode
 * @src: Data to be encoded
 * @len: Length of the data to be encoded
 * @out_len: Pointer to output length variable, or %NULL if not used
 * Returns: Allocated buffer of out_len bytes of encoded data,
 * or %NULL on failure
 *
 * Caller is responsible for freeing the returned buffer. Returned buffer is
 * nul terminated to make it easier to use as a C string. The nul terminator is
 * not included in out_len.
 */
unsigned char* b64_encode(const unsigned char *src, size_t len, size_t *out_len)
{
    unsigned char *out, *pos;
    const unsigned char *end, *in;
    size_t olen;
    int line_len;

    olen = len * 4 / 3 + 4; /* 3-byte blocks to 4-byte */
    olen += olen / 72; /* line feeds */
    olen++; /* nul termination */
    if (olen < len)
        return NULL; /* integer overflow */
    out = malloc(olen);
    if (out == NULL)
        return NULL;

    end = src + len;
    in = src;
    pos = out;
    line_len = 0;
    while (end - in >= 3) 
        {
        *pos++ = base64_table[in[0] >> 2];
        *pos++ = base64_table[((in[0] & 0x03) << 4) | (in[1] >> 4)];
        *pos++ = base64_table[((in[1] & 0x0f) << 2) | (in[2] >> 6)];
        *pos++ = base64_table[in[2] & 0x3f];
        in += 3;
        line_len += 4;
        if (line_len >= 72) 
            {
            *pos++ = '\n';
            line_len = 0;
            }
	}

    if (end - in) 
        {
        *pos++ = base64_table[in[0] >> 2];
        if (end - in == 1) 
            {
            *pos++ = base64_table[(in[0] & 0x03) << 4];
            *pos++ = '=';
            } 
        else 
            {
            *pos++ = base64_table[((in[0] & 0x03) << 4) |
					      (in[1] >> 4)];
            *pos++ = base64_table[(in[1] & 0x0f) << 2];
            }
        *pos++ = '=';
        line_len += 4;
        }

    if (line_len)
        *pos++ = '\n';

    *pos = '\0';
    if (out_len)
        *out_len = pos - out;
    return out;
}

/* read config */
void config_read (void)
{
    int i=0;
    char *tmp,line[16],buffer[16];
    ESP_LOGI(TAG, "Reading file");
    FILE* f = fopen("/spiffs/config.txt", "r");
    if (f == NULL) 
        {
        ESP_LOGE(TAG, "Failed to open file for reading");
        status = PORT_OFF;
        strcpy(config.name, WEBSERVER_ESP_STATION_NAME);
        strcpy(config.ssid, WEBSERVER_ESP_WIFI_SSID);
        strcpy(config.password, WEBSERVER_ESP_WIFI_PASS);
        strcpy(config.station_password, WEBSERVER_ESP_STATION_PASSWORD);
        strcpy(config.use_dhcp,"yes");
        ESP_LOGI(TAG, "Config read failed, trying to connect to SSID %s with password %s.",config.ssid, config.password);
        return;
        }
    fgets(line, sizeof(line), f);
    fgets(config.ssid, sizeof(config.ssid), f);
    fgets(config.password, sizeof(config.password), f);
    fgets(config.name, sizeof(config.name), f);
    fgets(config.ip, sizeof(config.ip), f);
    fgets(config.gw, sizeof(config.gw), f);
    fgets(config.netmask, sizeof(config.netmask), f);
    fgets(config.use_dhcp, sizeof(config.use_dhcp), f);
    fgets(config.station_password, sizeof(config.station_password), f);
    fgets(config.use_authorization, sizeof(config.use_authorization), f);
    fgets(config.dns1, sizeof(config.dns1), f);
    fgets(config.dns2, sizeof(config.dns2), f);
    fgets(config.auto_ota, sizeof(config.auto_ota), f);
    fgets(config.use_actions, sizeof(config.use_actions), f);
    fgets(config.use_mdns, sizeof(config.use_mdns), f);
    fgets(config.update_result, sizeof(config.update_result), f);
    fclose(f);
    // strip newline
    tmp = strchr(line, '\n');
    if (tmp) 
        *tmp = '\0';
    tmp = strchr(config.ssid, '\n');
    if (tmp) 
        *tmp = '\0';
    tmp = strchr(config.password, '\n');
    if (tmp) 
        *tmp = '\0';
   tmp = strchr(config.name, '\n');
    if (tmp) 
        *tmp = '\0';
    tmp = strchr(config.ip, '\n');
    if (tmp) 
        *tmp = '\0';
    tmp = strchr(config.gw, '\n');
    if (tmp) 
        *tmp = '\0';
    tmp = strchr(config.netmask, '\n');
    if (tmp) 
        *tmp = '\0';
    tmp = strchr(config.use_dhcp, '\n');
    if (tmp) 
        *tmp = '\0';
    tmp = strchr(config.station_password, '\n');
    if (tmp) 
        *tmp = '\0';
    tmp = strchr(config.use_authorization, '\n');
    if (tmp) 
        *tmp = '\0';
    tmp = strchr(config.dns1, '\n');
    if (tmp) 
        *tmp = '\0';
    tmp = strchr(config.dns2, '\n');
    if (tmp) 
        *tmp = '\0';
    tmp = strchr(config.auto_ota, '\n');
    if (tmp) 
        *tmp = '\0';
    tmp = strchr(config.use_actions, '\n');
    if (tmp) 
        *tmp = '\0';
    tmp = strchr(config.use_mdns, '\n');
    if (tmp) 
        *tmp = '\0';
    tmp = strchr(config.update_result, '\n');
    if (tmp) 
        *tmp = '\0';
    status = atoi (line);
    if (*config.ssid)
        ESP_LOGI(TAG, "Read from file: station name:%s,status:%i,SSID:%s,ip:%s,gw:%s,netmask:%s, DNS1:%s, DNS2:%s, use_dhcp:%s, use_authorization:%s, auto_ota:%s, use_actions:%s, use_mdns:%s", config.name,status,config.ssid,config.ip,config.gw,config.netmask,config.dns1,config.dns2,config.use_dhcp,config.use_authorization,config.auto_ota,config.use_actions,config.use_mdns);
    else
        {
        status = PORT_OFF;
        strcpy(config.name, WEBSERVER_ESP_STATION_NAME);
        strcpy(config.ssid, WEBSERVER_ESP_WIFI_SSID);
        strcpy(config.password, WEBSERVER_ESP_WIFI_PASS);
        strcpy(config.station_password, WEBSERVER_ESP_STATION_PASSWORD);
        strcpy(config.use_dhcp,"yes");
        strcpy(config.use_mdns,"yes");
        ESP_LOGI(TAG, "Config read failed, trying to connect to SSID %s with password %s.",config.ssid, config.password);
        }    
    
    FILE* f2 = fopen("/spiffs/actions.txt", "r");
    for (i=0;i<SWITCHLIST_SIZE;i++)
        {
        fgets(buffer,sizeof(buffer), f2);
        tmp = strchr(line, '\n');
        if (tmp) 
            *tmp = '\0';
        switch_times[i].event_status = atoi (buffer);
        fgets(buffer,sizeof(buffer), f2);
        tmp = strchr(line, '\n');
        if (tmp) 
            *tmp = '\0';
        switch_times[i].tm_min = atoi (buffer);
        fgets(buffer,sizeof(buffer), f2);
        tmp = strchr(line, '\n');
        if (tmp) 
            *tmp = '\0';
        switch_times[i].tm_hour = atoi (buffer);
        fgets(buffer,sizeof(buffer), f2);
        tmp = strchr(line, '\n');
        if (tmp) 
            *tmp = '\0';
        switch_times[i].tm_wday = atoi (buffer); 
        fgets(buffer,sizeof(buffer), f2);
        tmp = strchr(line, '\n');
        if (tmp) 
            *tmp = '\0';
        switch_times[i].tm_daynr = atoi (buffer); 
        fgets(buffer,sizeof(buffer), f2);
        tmp = strchr(line, '\n');
        if (tmp) 
            *tmp = '\0';
        switch_times[i].tm_monthnr = atoi (buffer); 
       fgets(buffer,sizeof(buffer), f2);
        tmp = strchr(line, '\n');
        if (tmp) 
            *tmp = '\0';
        switch_times[i].recurring = atoi (buffer); 
        fgets(buffer,sizeof(buffer), f2);
        tmp = strchr(line, '\n');
        if (tmp) 
            *tmp = '\0';
        switch_times[i].switch_status = atoi (buffer); 
        }
    fclose(f2);
}

/* write config */
void config_write (int fstatus)
{
    int i=0;
    ESP_LOGI(TAG, "Opening file");
    FILE* f = fopen("/spiffs/config.txt", "w");
    if (f == NULL) 
        {
        ESP_LOGE(TAG, "Failed to open file for writing");
        return;
        }
    fprintf(f, "%i\n",fstatus);
    fprintf(f, "%s\n",(char *)config.ssid);
    fprintf(f, "%s\n",(char *)config.password);
    fprintf(f, "%s\n",(char *)config.name);
    fprintf(f, "%s\n",(char *)config.ip);
    fprintf(f, "%s\n",(char *)config.gw);
    fprintf(f, "%s\n",(char *)config.netmask);
    fprintf(f, "%s\n",(char *)config.use_dhcp);
    fprintf(f, "%s\n",(char *)config.station_password);
    fprintf(f, "%s\n",(char *)config.use_authorization);
    fprintf(f, "%s\n",(char *)config.dns1);
    fprintf(f, "%s\n",(char *)config.dns2);
    fprintf(f, "%s\n",(char *)config.auto_ota);
    fprintf(f, "%s\n",(char *)config.use_actions);
    fprintf(f, "%s\n",(char *)config.use_mdns);
    fprintf(f, "%s\n",(char *)config.update_result);
    fclose(f);
    FILE* f2 = fopen("/spiffs/actions.txt", "w");
    for (i=0;i<SWITCHLIST_SIZE;i++)
        {
        fprintf(f2, "%i\n",switch_times[i].event_status);
        fprintf(f2, "%i\n",switch_times[i].tm_min);
        fprintf(f2, "%i\n",switch_times[i].tm_hour);
        fprintf(f2, "%i\n",switch_times[i].tm_wday);
        fprintf(f2, "%i\n",switch_times[i].tm_daynr);
        fprintf(f2, "%i\n",switch_times[i].tm_monthnr);
        fprintf(f2, "%i\n",switch_times[i].recurring);   
        fprintf(f2, "%i\n",switch_times[i].switch_status);    
        }
    fclose(f2);
    ESP_LOGI(TAG, "File written");
}

static void connect_handler(void* arg, esp_event_base_t event_base, int32_t event_id, void* event_data)
{
    httpd_handle_t* server = (httpd_handle_t*) arg;
    wifi_state = 1;
    if (*server == NULL) 
        {
        ESP_LOGI(TAG, "Starting webserver");
        *server = start_webserver();
        }
}

/*handlers for connect/disconnect */
static void disconnect_handler(void* arg, esp_event_base_t event_base, int32_t event_id, void* event_data)
{
    s_wifi_event_group = xEventGroupCreate();
    httpd_handle_t* server = (httpd_handle_t*) arg;
    wifi_state = 0;
    if (*server) 
        {
        ESP_LOGI(TAG, "Stopping webserver");
        stop_webserver(*server);
        *server = NULL;
        //webserver will be restarted by handler connect_handler again after wifi reconnect.
        }
  
}

/* event handler */
static void event_handler(void* arg, esp_event_base_t event_base,int32_t event_id, void* event_data)
{
    if (event_base == WIFI_EVENT && event_id == WIFI_EVENT_STA_START) 
        esp_wifi_connect();
    else if (event_base == WIFI_EVENT && event_id == WIFI_EVENT_STA_DISCONNECTED) 
        {
        if (s_retry_num < WEBSERVER_ESP_MAXIMUM_RETRY) 
            {
            esp_wifi_connect();
            s_retry_num++;
            ESP_LOGI(TAG, "retry to connect to the AP");
            } 
        else
            xEventGroupSetBits(s_wifi_event_group, WIFI_FAIL_BIT);
        ESP_LOGI(TAG,"connect to the AP fail");
        } 
    else if (event_base == IP_EVENT && event_id == IP_EVENT_STA_GOT_IP) 
        {
        ip_event_got_ip_t* event = (ip_event_got_ip_t*) event_data;
        ESP_LOGI(TAG, "got ip:%s",ip4addr_ntoa(&event->ip_info.ip));
        s_retry_num = 0;
        xEventGroupSetBits(s_wifi_event_group, WIFI_CONNECTED_BIT);
        }
}

esp_err_t http_client_event_handler(esp_http_client_event_t *evt)
{
    switch(evt->event_id) {
        case HTTP_EVENT_ERROR:
            ESP_LOGD(TAG, "HTTP_EVENT_ERROR");
            break;
        case HTTP_EVENT_ON_CONNECTED:
            ESP_LOGD(TAG, "HTTP_EVENT_ON_CONNECTED");
            break;
        case HTTP_EVENT_HEADER_SENT:
            ESP_LOGD(TAG, "HTTP_EVENT_HEADER_SENT");
            break;
        case HTTP_EVENT_ON_HEADER:
            ESP_LOGD(TAG, "HTTP_EVENT_ON_HEADER, key=%s, value=%s", evt->header_key, evt->header_value);
            break;
        case HTTP_EVENT_ON_DATA:
            ESP_LOGD(TAG, "HTTP_EVENT_ON_DATA, len=%d", evt->data_len);
            break;
        case HTTP_EVENT_ON_FINISH:
            ESP_LOGD(TAG, "HTTP_EVENT_ON_FINISH");
            break;
        case HTTP_EVENT_DISCONNECTED:
            ESP_LOGD(TAG, "HTTP_EVENT_DISCONNECTED");
            break;
    }
    return ESP_OK;
}

/* Converts a hex character to its integer value */
char from_hex(char ch) 
{
    return isdigit(ch) ? ch - '0' : tolower(ch) - 'a' + 10;
}

int isIp_v4(char* ip)
{
    int num, dots = 0;
    char* ptr;

    if (ip == NULL)
        return 0;

    ptr = strtok(ip, ".");

    if (ptr == NULL)
        return 0;

    while (ptr) {

        /* after parsing string, it must contain only digits */
        if (!valid_digit(ptr))
            return 0;

        num = atoi(ptr);

        /* check for valid IP */
        if (num >= 0 && num <= 255) {
            /* parse remaining string */
            ptr = strtok(NULL, ".");
            if (ptr != NULL)
                ++dots;
        }
        else
            return 0;
    }

    /* valid IP string must contain 3 dots */
    if (dots != 3)
        return 0;
    return 1;
}

int get_day_number(char *s)
{
    int weekday_tmp;
    if (strcmp(s,"Sunday") == 0)
        weekday_tmp=0;
    else if (strcmp(s,"Monday") == 0)
        weekday_tmp=1;
    else if (strcmp(s,"Tuesday") == 0)
        weekday_tmp=2;
    else if (strcmp(s,"Wednesday") == 0)
        weekday_tmp=3;
    else if (strcmp(s,"Thursday") == 0)
        weekday_tmp=4;
    else if (strcmp(s,"Friday") == 0)
        weekday_tmp=5;
    else if (strcmp(s,"Saturday") == 0)
        weekday_tmp=6;
    else if (strcmp(s,"Workdays") == 0)
        weekday_tmp=7;
    else if (strcmp(s,"Weekend") == 0)
        weekday_tmp=8;
    else
        weekday_tmp=9;
    return weekday_tmp;
}

int get_month_number(char *s)
{
    int monthnr_tmp = 0;
    if (strcmp(s,"Jan") == 0)
        monthnr_tmp=1;
    else if (strcmp(s,"Feb") == 0)
        monthnr_tmp=2;
    else if (strcmp(s,"Mar") == 0)
        monthnr_tmp=3;
    else if (strcmp(s,"Apr") == 0)
        monthnr_tmp=4;
    else if (strcmp(s,"Mai") == 0)
        monthnr_tmp=5;
    else if (strcmp(s,"June") == 0)
        monthnr_tmp=6;
    else if (strcmp(s,"July") == 0)
        monthnr_tmp=7;
    else if (strcmp(s,"Aug") == 0)
        monthnr_tmp=8;
    else if (strcmp(s,"Sep") == 0)
        monthnr_tmp=9;
    else if (strcmp(s,"Oct") == 0)
        monthnr_tmp=10; 
    else if (strcmp(s,"Nov") == 0)
        monthnr_tmp=11;
    else if (strcmp(s,"Dec") == 0)
        monthnr_tmp=12;
    else
        monthnr_tmp=14;
    return monthnr_tmp;
}

/* obtain time */
static void obtain_time(void)
{
    // wait for time to be set
    time_t now = 0;
    struct tm timeinfo = { 0 };
    int retry = 0;
    const int retry_count = 10;

    while (timeinfo.tm_year < (2016 - 1900) && ++retry < retry_count) 
        {
        ESP_LOGI(TAG, "Waiting for system time to be set... (%d/%d)", retry, retry_count);
        vTaskDelay(2000 / portTICK_PERIOD_MS);
        time(&now);
        localtime_r(&now, &timeinfo);
        }
} 

//void ota_task(void * pvParameter)
void ota_updater(void)
{
    int i;
    ESP_LOGI(TAG, "Starting OTA...");
    
    esp_http_client_config_t  http_client_config = {
        .url = OTA_SERVER_URL,
        .cert_pem = OTA_SERVER_CERT,
        .event_handler = http_client_event_handler,
        };
    for (i=0;i<UPDATE_FAIL_RETRY;i++)
        {
        esp_err_t ret = esp_https_ota(&http_client_config);
        if (ret == ESP_OK)
            {
            ESP_LOGI(TAG, "Firmware Upgrade done!");
            strcpy(config.auto_ota,"no");
            strcpy(config.update_result,UPDATE_SUCCESS);
            config_write(status);
            esp_restart();
            return;
            }
        else
            {
            ESP_LOGE(TAG, "Firmware Upgrade failed!");
            strcpy(config.auto_ota,"no");
            strcpy(config.update_result,UPDATE_FAILED);
            config_write(status);
            }
        }
}

// toggle output port
void port_switch(int port_status)
{
    if (next_switch_delay > 0 || port_status == status)
        return;
    status = port_status;
    /*update time */
    time(&now);
    localtime_r(&now, &timeinfo);
    strftime(strftime_buf, sizeof(strftime_buf), "%c", &timeinfo);
    if (port_status == PORT_ON)
        {
        strcat(strftime_buf," / on(web)!");
        ESP_LOGI(TAG, "Toggling PORT: on");
        }
    else if (port_status == PORT_OFF)
        {
        strcat(strftime_buf," / off(web)!");
        /*toggle Port */
        ESP_LOGI(TAG, "Toggling PORT: off");
        }
    else
        {
        strcat(strftime_buf," / timer(web)!");
        /*toggle Port */
        ESP_LOGI(TAG, "Toggling PORT: on/timer");
        }
    next_switch_delay = 50; // wait a bit until next switch is allowed
    gpio_set_level(GPIO_OUTPUT_IO_0, port_status); 
    generate_response(port_status);
    if (port_status == PORT_TIMER)
        port_status = PORT_ON;
    config_write (port_status);
}

//Helper for action list sorting, Hierarchy recurring->monthnr->daynr->hour->min
int qsort_helper( const void *left, const void *right )
{
    const time_struct *a = ( const time_struct *)left;
    const time_struct *b = ( const time_struct *)right;

    if ( b->recurring < a->recurring )
    {
        return -1;
    }
    else if ( a->recurring < b->recurring )
    {
        return 1;
    }
    if ( b->tm_monthnr < a->tm_monthnr )
    {
        return 1;
    }
    else if ( a->tm_monthnr < b->tm_monthnr )
    {
        return -1;
    }
    if ( b->tm_daynr < a->tm_daynr )
    {
        return 1;
    }
    else if ( a->tm_daynr < b->tm_daynr )
    {
        return -1;
    }
    if ( b->tm_wday < a->tm_wday )
    {
        return 1;
    }
    else if ( a->tm_wday < b->tm_wday )
    {
        return -1;
    }    
    if ( b->tm_hour < a->tm_hour )
    {
        return 1;
    }
    else if ( a->tm_hour < b->tm_hour )
    {
        return -1;
    }
    else
    {
        return  (  a->tm_min > b->tm_min ) - ( b->tm_min > a->tm_min );
    }
}

//Check if we have a planned switch event. Do switch and disable events that should only occur once (tm.wday = 7).
void switch_checker(void)
{
    int i = 0, do_switch = 0;
    time(&now);
    localtime_r(&now, &timeinfo);
    strftime(strftime_buf, sizeof(strftime_buf), "%c", &timeinfo);
    for (i=0;i<SWITCHLIST_SIZE;i++)
        {
        //Lets first check if we have a planned event for actual time,if its active and if the port status is different from actual status.
        if (switch_times[i].event_status && switch_times[i].tm_min == timeinfo.tm_min && switch_times[i].tm_hour == timeinfo.tm_hour && switch_times[i].switch_status != status)
            {
            //Test for a one-time event
            if (switch_times[i].recurring == RECUR_ONCE && switch_times[i].tm_monthnr == timeinfo.tm_mon+1 && switch_times[i].tm_daynr == timeinfo.tm_mday )
                {
                do_switch = 1;
                switch_times[i].event_status = 0;
                }
            //Test for a yearly event
            else if (switch_times[i].recurring == RECUR_YEAR && switch_times[i].tm_monthnr == timeinfo.tm_mon+1 && switch_times[i].tm_daynr == timeinfo.tm_mday )
                do_switch = 1;
            //Test for a monthly event
            else if (switch_times[i].recurring == RECUR_MONTH && switch_times[i].tm_daynr == timeinfo.tm_mday)
                do_switch = 1;
            //Test for a weekly event
            //This statement tests the weekday, there's the check if the actual day matches exactly or if we are within weekend or weekdays.
            else if (switch_times[i].recurring == RECUR_WEEK && (switch_times[i].tm_wday == timeinfo.tm_wday || (switch_times[i].tm_wday == 7 && (timeinfo.tm_wday > 0 && timeinfo.tm_wday < 6))  || (switch_times[i].tm_wday == 8 && (timeinfo.tm_wday == 0 || timeinfo.tm_wday == 6))))
                do_switch = 1;
            //Test for a all-day-event
            else if (switch_times[i].recurring == RECUR_DAY)
                do_switch = 1;
            }
        if (do_switch)
            {
            //enable blocking of of switch functions for 61 secs to avoid multiple executions. 61 secs because with 60 secs in can happen that with two actions 
            //which should occur at the same time point, both actions will be executed instead of only the higher one as expected.
            switch_action_done = 1;
            switch_action_timer = 61;
            //Port switch
            ESP_LOGI(TAG, "Time: %s Timer Event detected! Switched Port to status %i", strftime_buf, switch_times[i].switch_status);
            port_switch(switch_times[i].switch_status);
            break;
            }
        }
}
    
/* Returns a url-decoded version of str */
/* IMPORTANT: be sure to free() the returned string after use */
char *url_decode(char *str) 
    {
    char *pstr = str, *buf = malloc(strlen(str) + 1), *pbuf = buf;
    while (*pstr) 
        {
        if (*pstr == '%') 
            {
            if (pstr[1] && pstr[2]) 
                {
                *pbuf++ = from_hex(pstr[1]) << 4 | from_hex(pstr[2]);
                pstr += 2;
                }
            } 
        else if (*pstr == '+') 
            { 
            *pbuf++ = ' ';
            } 
        else 
            {
            *pbuf++ = *pstr;
            }
        pstr++;
        }
    *pbuf = '\0';
    return buf;
}

int valid_digit(char* ip_str)
{
    while (*ip_str) {
        if (*ip_str >= '0' && *ip_str <= '9')
            ++ip_str;
        else
            return 0;
    }
    return 1;
}
